# ========================================================================
'''                 CSDMS AI&ML CHALLENGE DATASET_v1.py

Started 19 May 2019
Version: 1 19 May 2019
Author: CJ Jenkins INSTAAR

Program to load dbSEABED gridded data into memory and support
         the CSDMS AI&ML CHALLENGE. A possible framework
         for people attempting the challenge.

i. Loads in the dbSEABED gridded sample data (training)
ii. Loads in the Feature layers (predictors)

Some abbreviations...
   pj* - project*
   *L - a list; *A - an array; *S - a string; *D - a dictionary; *T - tuple



***************************************************************************
Legal notice...
This software is research grade, a workbench. While every effort has been made
to achieve high quality and reliability it should not be used in any
endeavour supporting decision-making in policy, investment, engineering,
navigation, or any other field.
If that is required, please contact the University of Colorado, Boulder
to obtain an approved fit-for-purpose version.
***************************************************************************


'''
# ========================================================================
#Banner ----------------------------------------------------
print "*"*75; print
print "Program accompanying the CSDMS AI&ML CHALLENGE DATASET"
print; print "*"*75
print
print "   (If a delay, loading special Python modules...)"
#-----------------------------------------------------------

import glob,os,time
import numpy as NPY
import matplotlib.pyplot as PLT

from sklearn.ensemble import RandomForestClassifier
from sklearn.ensemble import RandomForestRegressor

from scipy import ndimage as NDI

projS="atlUsa3"
#It is assumed that this program and all files will be in the same Folder

fNaN=float('NaN')
loudQ=False #..determines whether program reports verbosely

try: os.mkdir("_Graphics")
except: pass #... probably already exists

# =================================================================
def performMachineLearning():

    print;print ">>performMachineLearning:"

    #INSERT YOUR METHOD HERE

    return

# =================================================================
def rfFeatureSelection(X_train,y_train):

    #NOT OPERATIONAL YET

    print;print ">>rfFeatureSelection: X_cols,y_col=",X_cols,y_col
    print X_train.shape,minMax(X_train),NPY.sum(X_train)
    print y_train.shape,minMax(y_train),NPY.sum(y_train)
    #...X_cols,y_col in from global

    from sklearn.feature_selection import SelectFromModel

    #See: "https://towardsdatascience.com/feature-selection-using-random-forest-26d7b747597f"

    sel = SelectFromModel(RandomForestClassifier(n_estimators = 100))
    sel.fit(X_train, y_train)

    supportdColsA=sel.get_support()
    #It will return an array of boolean values. True for the features whose importance is greater than the mean importance and False for the rest.

    #5. We can now make a list and count the selected features.

    selectedFeatsL=[X_c for cI,X_c in enumerate(X_cols)
                    if supportdColsA[cI]]

    print "selectedFeatsL=",selectedFeatsL

    return

# =================================================================
def uniquifyFeatures():

    #NOT FULLY OPERATIONAL YET

    #This gives much greater efficiency in the application of the fitted ML
    #   to large spatial griddings, especially when resolution is high.

    # ------------------------------------------------------
    #Get the unique sets from the feature value-sets
    # they will be assigned a serial number token ...
    
    print;print "Obtaining the unique, non-NaN feature entries..."
    featrUniqueSetsL=[]
    uniqFeatrRowsL=list(set([tuple(row) \
                      for row in trainDataValuesA[1:]\
                      if ~NPY.isnan(sum(row))]))
    #...with the [1:] we cut off the geological sample information.
    print "trainDataValuesA.shape,len(uniqFeatrRowsL)=",\
          trainDataValuesA.shape,len(uniqFeatrRowsL)

    uniqFeatrRowIdxsL=range(len(uniqFeatrRowsL))

    #Assign their indexes (number codes) ----------------------
    print;print "Assigning numeric index to each unique enviro ..."
    print "    (Patience... long process over whole map area.)"
    featrUniqIndxsMapA=NPY.ones_like(dpthProjA)*fNaN #...initializing
    print "featrUniqIndxsMapA.shape=",featrUniqIndxsMapA.shape

    #Build one BIG stack of feature layers -------------------------  
    totalFeatrStackA=NPY.dstack([dpthProjA,slpProjA,aspProjA,\
            proxProjA,bpiProjA,triProjA,hsnProjA,\
            btmpProjA,stmpProjA,bCurrProjA,\
            ubWw3ProjA,kd490ProjA,chloraProjA])
    #... arrays of 13 elements
    print "totalFeatrStackA.shape=",totalFeatrStackA.shape,\
          totalFeatrStackA[:,:,0].shape

    #Find the uniques in that stack using 'set' on tuples
    #   of each column, (for memory limits) per (longitude) row 
    #   then (latitude-wise) by column down the whole stack, finally also
    #   using 'set' to weed out duplicates in the resulting
    #   list of uniques ...
    featrUniqueSetsL=[]
    #..initialize to take the unique verticals in the stack...
    for row in totalFeatrStackA:
        rowUniqs=list(set([tuple(col) for col in row\
                      if ~NPY.isnan(sum(col))]))
        featrUniqueSetsL.extend(rowUniqs)
        #...all the tuples are written separately
    featrUniqueSetsL=list(set(featrUniqueSetsL))

    if loudQ:
        print "Orig: len(featrUniqueSetsL),featrUniqueSetsL[]=",\
              len(featrUniqueSetsL),featrUniqueSetsL[:4],"...",\
              len(featrUniqueSetsL[0])
        #... may be some NaN's in it

    #Clean any that have a NaN enviro value...
    featrUniqueSetsL=[eUS for eUS in featrUniqueSetsL\
                    if ~NPY.isnan(sum(eUS))]

    if loudQ:
        print "Clean: len(featrUniqueSetsL),featrUniqueSetsL[]=",\
              len(featrUniqueSetsL),featrUniqueSetsL[:4],"...",\
              len(featrUniqueSetsL[0])
        #...This is the complete matrix of the enviro values, uniqui-fied

    #Assign their indexes (number codes) ----------------------
    print "Assigning numeric index to each unique enviro..."
    featrUniqIndxsMapA=NPY.ones_like(dpthProjA)*fNaN
    print "featrUniqIndxsMapA.shape=",featrUniqIndxsMapA.shape
    
    for sUi,sUe in enumerate(featrUniqueSetsL):

        uniqCoords=NPY.where((totalFeatrStackA==sUe).all(axis=2))
        #print "   uniqCoords=",uniqCoords
        featrUniqIndxsMapA[uniqCoords]=sUi

        if sUi % 1000==0:
            print "At sUi=",sUi," of ",len(featrUniqueSetsL)

    print "featrUniqIndxsMapA.shape=",featrUniqIndxsMapA.shape
        
    #Plot the "uniqIndxsMapA" grids ....
    featrUniqIndxsMapA=NPY.where(NPY.isnan(wdProjA),fNaN,featrUniqIndxsMapA)
    #...Land mask
    PLT.imshow(featrUniqIndxsMapA,interpolation='None',cmap="Paired")
    PLT.title(gridInS+"_iniFeatrUniquesMapA")
    cbar=PLT.colorbar(shrink=0.5)
    cbar.set_label("Env Unique Num",rotation=90,fontsize=10)
    #PLT.imshow(land1MaskA,interpolation='None',cmap='gray')
    #...the land mask
    PLT.savefig("_Graphics/"+gridInS+"_iniFeatrUniquesMap.png")
    PLT.close()
    
    print "End of uniquifying: len(envUniqueSetsL)=",len(envUniqueSetsL)
        
    return

# =================================================================
def holdBackTrainData():

    global trainDataValuesA
    global holdBackFeatrValuesA

    #Extracts 10% of the data, if requested, and returns the
    #   90% under old name, and 10% as a new array.

    dataLngth=trainDataValuesA.shape[0]

    print;print ">>holdBackTrainData: dataLngth=",dataLngth

    holdBackA=NPY.random.choice(range(dataLngth),dataLngth/10)
    maintainA=NPY.isin(range(dataLngth), holdBackA, invert=True)

    holdBackFeatrValuesA=trainDataValuesA[holdBackA]
    trainDataValuesA=trainDataValuesA[maintainA]

    print "trainDataValuesA.shape[0],holdBackFeatrValuesA.shape[0]=",\
          trainDataValuesA.shape[0],holdBackFeatrValuesA.shape[0]

    return

# =================================================================
def ingestValuesGrid(gridToGetS):

    print;print ">>ingestValuesGrid:"
    print "Getting input parameter data (ascDatMapA)..."
    print "       from: ",projS+gridToGetS+".asc"

    #Read in the ESRI ASCII data file ------------------------
    dtaHdrL=[]
    ascF=open(projS+gridToGetS+".asc","r")
    #...CJJ Checked 15Feb2018 - the ASC grid is right-way-up in GIS.
    for iH in range(6):
        ascHdrS=ascF.readline().strip() #Header
        while "  " in ascHdrS: ascHdrS=ascHdrS.replace("  "," ")
        dtaHdrL.append(ascHdrS)
    dtaHdrL=[aHA.split(" ") for aHA in dtaHdrL]
    ascNoDataVal=float(dtaHdrL[5][1])
    print "dtaHdrL=",dtaHdrL,ascNoDataVal

    #Set the basic grid parameters (dimensions) for the wholework ...
    dtaHdrDatL=[aHA[1] for aHA in dtaHdrL]
    print "dtaHdrDatL=",dtaHdrDatL
    dtaGridCoRoL=map(int,dtaHdrDatL[:2])
    dtaCellSz=float(dtaHdrDatL[4])
    dtaLlLoLaSz=map(float,dtaHdrDatL[2:4])
    dtaOrLoLaL=[dtaLlLoLaSz[0],\
                 dtaLlLoLaSz[1]+dtaCellSz*float(dtaGridCoRoL[1])]
    print "dtaGridCoRoL,dtaCellSz,dtaLlLoLaSz,dtaOrLoLaL=",\
          dtaGridCoRoL,dtaCellSz,dtaLlLoLaSz,dtaOrLoLaL

    #Read in the data itself...
    ascDatMapA=ascF.read().strip().split("\n") #...Gulp it in !
    ascDatMapA=[map(float,aDA.strip().split(" ")) for aDA in ascDatMapA]
    #For some parameters int is suitable (rck, etc),
    #    but not for others (srt, etc); so ...
    if gridToGetS in ["rck","gvl","snd","mud","crb","por"]:
        ascDatMapA=NPY.round(ascDatMapA,0)
    elif gridToGetS in ["ocbn"]:
        ascDatMapA=NPY.round(ascDatMapA,1)
    elif gridToGetS in ["grz","srt"]:
        ascDatMapA=NPY.round(ascDatMapA,2)
    else:
        ascDatMapA=NPY.array(ascDatMapA)

    print "ascDatMapA.shape|minMax=",\
          ascDatMapA.shape,minMax(ascDatMapA[ascDatMapA<>-99.])
    #...note that the -99. is preserved as Null !!
    ascF.close()

    return ascDatMapA,ascNoDataVal

# =================================================================
def makeTrainData():

    global ascNoDataVal
    global trainDataValuesA, trainDataLocnsT

    #print;print "reBuilding cells in 'PROJ_rfTrain.csv' ..."

    #Get an ASC grid, merge it with the terrain data,
    #   and that is our 'train' set

    #...Equivalent is buildGsmTrainingSet for G:S:M simplexes;
    #    this code is single values only like %carb.
        
    datSourceS="_Cell"
    #...type of submitted values data (can also be a partial interpln)
    print;print ">>makeTrainData: gridInS,datSourceS=",gridInS,datSourceS

    #Find all the cells where there are data values -------
    
    #First import the basic data ...
    ascDatMapA,ascNoDataVal=ingestValuesGrid(datSourceS+"_"+gridInS)

    #Collect all the actual data values  ...
    trainDataLocnsT=tuple(NPY.where(ascDatMapA<>ascNoDataVal))
    #... this is all the values, not just the uniques.
    print "len(trainDataLocnsT[0])|minMax=",len(trainDataLocnsT[0]),\
          minMax(trainDataLocnsT[0]),minMax(trainDataLocnsT[1])
                   
    trainDataValuesT=tuple(ascDatMapA[trainDataLocnsT])
    #...single values !
    #...No further changes are allowed to trainDataValuesT !

    trainDataValuesL=[ascDatMapA[trainDataLocnsT]]
    #... to anchor the main training set
    print "trainDataValuesL[0].shape|minMax(trainDataValuesL)=",\
          trainDataValuesL[0].shape,minMax(trainDataValuesL)
    
    #Add the environmental (feature) data to them ------------------------
    trainDataValuesL.append(dpthProjA[trainDataLocnsT])
    trainDataValuesL.append(slpProjA[trainDataLocnsT])
    trainDataValuesL.append(aspProjA[trainDataLocnsT])
    trainDataValuesL.append(proxProjA[trainDataLocnsT])
    trainDataValuesL.append(bpiProjA[trainDataLocnsT])
    trainDataValuesL.append(triProjA[trainDataLocnsT])
    trainDataValuesL.append(hsnProjA[trainDataLocnsT])
    trainDataValuesL.append(btmpProjA[trainDataLocnsT])
    trainDataValuesL.append(stmpProjA[trainDataLocnsT])
    trainDataValuesL.append(bCurrProjA[trainDataLocnsT])
    trainDataValuesL.append(ubWw3ProjA[trainDataLocnsT])
    trainDataValuesL.append(kd490ProjA[trainDataLocnsT])
    trainDataValuesL.append(chloraProjA[trainDataLocnsT])

    #for rowI,row in enumerate(trainDataValuesL):
    #    print "rowI,row.shape,row[:4]=",rowI,row.shape,row[:4]

    trainDataValuesA=NPY.array(trainDataValuesL).T
    if loudQ:
        print "trainDataValuesA[]=",trainDataValuesA[:4]
    del trainDataValuesL #...clear it out from memory

    #Filter against NaN's ...
    trainDataValuesA=NPY.array([tDV for tDV in trainDataValuesA\
                        if ~NPY.isnan(NPY.sum(tDV))])
    print "Final: trainDataValuesA.shape=",trainDataValuesA.shape

    X_train=trainDataValuesA[:,1:] #... enviro data columns
    y_train=trainDataValuesA[:,0] #...geo data column

    return X_train,y_train

# =================================================================
def getEnviroVariables():

    global dpthProjA,slpProjA,aspProjA,\
              proxProjA,hsnProjA,btmpProjA,stmpProjA
    global triProjA,bpiProjA
    global kd490ProjA,chloraProjA,rivrsProjA
    global bCurrProjA,ubWw3ProjA

    global pjCoRoA,pjCellSz,pjLlLoLaA,pjOrLoLaA

    global featureFilesL,featureDetailsL
    global X_cols,y_col
    
    loudQ=False
    print;print "-"*40
    print ">>getEnviroVariables: Get the basic (terrain, physical) variable grids"

    #Draw on terrain maps at least initially to classify an area
    #for rock, gravel, sand, etc.

    #These are the ones chosen to work with ...
    featureFilesL=[projS+"_elvnProj.asc",\
              projS+"_slpProj.asc",\
              projS+"_aspProj.asc",\
              projS+"_proxProj.asc",\
              projS+"_hsnProj.asc",\
              projS+"_btmpProj.asc",\
              projS+"_stmpProj.asc",\
              projS+"_bCurrProj.asc",\
              projS+"_ubWw3Proj.asc",\
              projS+"_bpiProj.asc",\
              projS+"_triProj.asc",\
              projS+"_kd490Proj.asc",\
              projS+"_chloraProj.asc"\
              ]

    featureDetailsL=["Water Depth (lg2,m)",\
                   "Slope (dg)",\
                   "Aspect (dg)",\
                   "Land proximity (km)",\
                   "Hessian Matrix lrg eignv (lg10 /10)",\
                   "Botm Temps (dgC)",\
                   "Srfc Temps (dgC)",\
                   "Botm Curr (m/s)",\
                   "Waves Ubmax (m/s)",\
                   "Bathym Posn Indx (m)",\
                   "Terrain Rghns Indx (m)",\
                   "Turbidity (kd490,m-1)",\
                   "Chlorophyll-a (mg/m3)"\
                     ]
    #Set the training data column headings...
    X_cols=[] #...will be appended to
    y_col=gridInS #...from beginning of the program

    print
    for tI,tFileS in enumerate(featureFilesL):

        print;print "Getting: tFileS=",tFileS

        tF=open(tFileS,"r")
        #... note that all files are in same Folder

        #Read in the Headers ...
        tHdrA=[]
        for tN in range(6):
            tHdrA.append(tF.readline())
        if loudQ: print "tHdrA=",tHdrA

        if tI==0:
            #First file to be read gives params ...
            pjHdrDatA=[tHA.strip().split()[1] for tHA in tHdrA]
            if loudQ: print "pjHdrDatA=",pjHdrDatA
            pjCoRoA=map(int,pjHdrDatA[:2])
            pjCellSz=float(pjHdrDatA[4])
            pjLlLoLaA=map(float,pjHdrDatA[2:4])
            pjOrLoLaA=[pjLlLoLaA[0],\
                         pjLlLoLaA[1]+pjCellSz*float(pjCoRoA[1])]
            if ~loudQ:
                print "pjCoRoA,pjCellSz,pjLlLoLaA,pjOrLoLaA=",\
                      pjCoRoA,pjCellSz,pjLlLoLaA,pjOrLoLaA
            #time.sleep(1)

        #Read in the Data ...
        tDatS=tF.read().strip()
        tDatA=tDatS.split("\n")
        tDatA=NPY.array([map(float,tDA.strip().split(" ")) for tDA in tDatA])

        featFileS=str(tFileS) #...initial name (may be changed)

        # --------------------------------------------------
        #Treat each one (they can potentially need different treatment) ...

        #It is important to continue to store the areal grids because
        #   they will be used in the application of the fitted ML model
        #   to get the final outputs ...
        
        if "elvnProj" in tFileS:
            wdProjA=NPY.where(tDatA>0.0,fNaN,-tDatA)
            #...set land to NaN; sea to Wd's
            wdProjA=NPY.where(wdProjA<0.1,0.1,wdProjA)
            #...adjust all that are close to zero
            #print NPY.unique(wdProjA[~NPY.isnan(wdProjA)])
            dpthProjA=NPY.log2(wdProjA)
            #... we will use log2 rescaling of WD's
            #...tDatA Usually -6538.... 596; this deals with land and zeros
            dpthProjA=NPY.where(NPY.isinf(dpthProjA),fNaN,dpthProjA)
            #...this clears out any infs (makes them NaN's)
            
            dpthProjA=NDI.median_filter(dpthProjA,3)
            #... smooths the input as an efficiency and anti-noise measure
            dpthProjA=norm4rf(dpthProjA,8)
            #print "NPY.unique(dpthProjA)=",NPY.unique(dpthProjA)
            #    log2 produces range 0.0...13.0
            #    There should be no nulls in this dataset...
            #...NaN's take care of themselves from here

            featFileS=projS+"_dpthProj" #... name change

            print "dpthProjA,minMax(tDatA)=",dpthProjA.shape,minMax(tDatA)
            print "   mnmx: minMax(dpthProjA)=",minMax(dpthProjA)
            print "   uniques: NPY.unique(dpthProjA).shape=",\
                  NPY.unique(dpthProjA[~NPY.isnan(dpthProjA)]).shape,\
                  NPY.unique(dpthProjA[~NPY.isnan(dpthProjA)])
            toPlotA=NPY.copy(dpthProjA)
            
        if "slpProj" in tFileS:
            slpProjA=NPY.round(tDatA,3) 
            #...tDatA Usually <=36
            #       There should be no nulls in this dataset...

            slpProjA=NDI.median_filter(slpProjA,3)
            #... smooths the input as an efficiency and anti-noise measure
            slpProjA=NPY.where(NPY.isnan(wdProjA),fNaN,slpProjA)

            slpProjA=norm4rf(slpProjA,8) #...normalize
            print "slpProjA,minMax(tDatA)=",slpProjA.shape,minMax(tDatA)
            print "   mnmx: minMax(slpProjA)=",minMax(slpProjA)
            print "   uniques: NPY.unique(slpProjA).shape=",\
                  NPY.unique(slpProjA[~NPY.isnan(slpProjA)]).shape
            toPlotA=NPY.copy(slpProjA)

        if "aspProj" in tFileS:
            aspProjA=NPY.round(tDatA,3) 
            #...tDatA Usually <=360
            #       There should be no nulls in this dataset...

            aspProjA=NDI.median_filter(aspProjA,3)
            #... smooths the input as an efficiency and anti-noise measure
            aspProjA=NPY.where(NPY.isnan(wdProjA),fNaN,aspProjA)

            aspProjA=norm4rf(aspProjA,8) #...normalize
            print "aspProjA,minMax(tDatA)=",aspProjA.shape,minMax(tDatA)
            print "   mnmx: minMax(aspProjA)=",minMax(aspProjA)
            print "   uniques: NPY.unique(aspProjA).shape=",\
                  NPY.unique(aspProjA[~NPY.isnan(aspProjA)]).shape
            toPlotA=NPY.copy(aspProjA)

        if "proxProj" in tFileS:
            #There are nulls in this dataset...
            proxProjA=NPY.where(tDatA<0.,fNaN,NPY.round(tDatA,3))
            #...tDatA Usually 0...51
            
            #Doesn't need denoising ...
            proxProjA=norm4rf(proxProjA,8) #...normalize
            print "proxProjA,minMax(tDatA)=",proxProjA.shape,minMax(tDatA)
            print "   mnmx: minMax(proxProjA)=",minMax(proxProjA)
            print "   uniques: NPY.unique(proxProjA).shape=",\
                  NPY.unique(proxProjA[~NPY.isnan(proxProjA)]).shape
            toPlotA=NPY.copy(proxProjA)
            
        if "bpiProj" in tFileS:
            bpiProjA=NPY.round(tDatA,0) #...accepts it
            print "minMax(bpiProjA)=",minMax(bpiProjA)

            bpiProjA=NDI.median_filter(bpiProjA,3)
            #... smooths the input as an efficiency and anti-noise measure
            bpiProjA=NPY.where(NPY.isnan(wdProjA),fNaN,bpiProjA)
            print "minMax(bpiProjA)=",minMax(bpiProjA)

            bpiProjA=norm4rf(bpiProjA,8) #...normalize
            print "bpiProj: minMax(tDatA)=",minMax(tDatA)
            print "   mnmx: minMax(bpiProjA)=",minMax(bpiProjA)
            print "   uniques: NPY.unique(bpiProjA).shape=",\
                  NPY.unique(bpiProjA[~NPY.isnan(bpiProjA)]).shape
            toPlotA=NPY.copy(bpiProjA)

        if "triProj" in tFileS:
            triProjA=NPY.round(tDatA,3) #...accepts it
            print "minMax(triProjA)=",minMax(triProjA)

            triProjA=NDI.median_filter(triProjA,3)
            #... smooths the input as an efficiency and anti-noise measure
            triProjA=NPY.where(NPY.isnan(wdProjA),fNaN,triProjA)
            print "minMax(triProjA)=",minMax(triProjA)

            triProjA=norm4rf(triProjA,8) #...normalize
            print "triProj: minMax(tDatA)=",minMax(tDatA)
            print "   mnmx: minMax(triProjA)=",minMax(triProjA)
            print "   uniques: NPY.unique(triProjA).shape=",\
                  NPY.unique(triProjA[~NPY.isnan(triProjA)]).shape
            toPlotA=NPY.copy(triProjA)
        

        if "hsnProj" in tFileS:
            #Clip it ...
            hsnProjA=NPY.where(tDatA<-100.,-100.,tDatA)
            hsnProjA=NPY.where(hsnProjA>+100.,+100.,hsnProjA)
            #...tDatA Usually -65....232.389
            #...There should be no nulls in this dataset.

            hsnProjA=NDI.median_filter(hsnProjA,3)
            #... smooths the input as an efficiency and anti-noise measure
            hsnProjA=NPY.where(NPY.isnan(wdProjA),fNaN,hsnProjA)

            hsnProjA=norm4rf(hsnProjA,8) #...normalize
            print "hsnProjA,minMax(tDatA)=",hsnProjA.shape,minMax(tDatA)
            print "   mnmx: minMax(hsnProjA)=",minMax(hsnProjA)
            print "   uniques: NPY.unique(hsnProjA).shape=",\
                  NPY.unique(hsnProjA[~NPY.isnan(hsnProjA)]).shape
            toPlotA=NPY.copy(hsnProjA)
            
        if "btmpProj" in tFileS:
            btmpProjA=NPY.round(tDatA,3) #...accepts it
            #...tDatA Usually -3....30
            #... reduced the value-range to ~0..~10
            #   There should be no nulls in this dataset.
            #Note: there is a striping in this dataset - CJJ to fix

            #Doesn't need denoising ...

            btmpProjA=norm4rf(btmpProjA,8) #...normalize
            print "btmpProjA,minMax(tDatA)=",btmpProjA.shape,minMax(tDatA)
            print "   mnmx: minMax(btmpProjA)=",minMax(btmpProjA)
            print "   uniques: NPY.unique(btmpProjA).shape=",\
                  NPY.unique(btmpProjA[~NPY.isnan(btmpProjA)]).shape
            toPlotA=NPY.copy(btmpProjA)

        if "stmpProj" in tFileS:
            stmpProjA=NPY.round(NPY.where(tDatA==-9999.,fNaN,tDatA),3) #...accepts it
            #...tDatA Usually -3....30 but can be 23..26
            #... reduced the value-range to ~0..~10
            #There should be no nulls in this dataset...
            #Note: there is a stiping in this dataset - CJJ to fix

            #Doesn't need denoising ...
            stmpProjA=norm4rf(stmpProjA,4) #...normalize
            print "stmpProjA,minMax(tDatA)=",stmpProjA.shape,minMax(tDatA)
            print "   mnmx: minMax(stmpProjA)=",minMax(stmpProjA)
            print "   uniques: NPY.unique(stmpProjA).shape=",\
                  NPY.unique(stmpProjA[~NPY.isnan(stmpProjA)]).shape
            toPlotA=NPY.copy(stmpProjA)

        if "bCurrProj" in tFileS:
            bCurrProjA=NPY.round(NPY.where(tDatA==-9999.,fNaN,tDatA),2) #...accepts it
            #...tDatA Usually -3....30 but can be 23..26
            #There should be no nulls in this dataset...
            #Note: there is a striping in this dataset - CJJ to fix

            #Doesn't need denoising ...
            bCurrProjA=NPY.where(NPY.isnan(wdProjA),fNaN,bCurrProjA)

            bCurrProjA=norm4rf(bCurrProjA,4) #...normalize
            print "bCurrProjA,minMax(tDatA)=",bCurrProjA.shape,minMax(tDatA)
            print "   mnmx: minMax(bCurrProjA)=",minMax(bCurrProjA)
            print "   uniques: NPY.unique(bCurrProjA).shape=",\
                  NPY.unique(bCurrProjA[~NPY.isnan(bCurrProjA)]).shape
            toPlotA=NPY.copy(bCurrProjA)
            
        if "ubWw3Proj" in tFileS:
            ubWw3ProjA=NPY.round(NPY.where(tDatA==-9999.,fNaN,tDatA),2) #...accepts it
            #...tDatA Usually 0....5 but can be 0..10

            #Doesn't need denoising ...
            ubWw3ProjA=NPY.where(NPY.isnan(wdProjA),fNaN,ubWw3ProjA)
            
            ubWw3ProjA=norm4rf(ubWw3ProjA,4) #...normalize
            print "ubWw3ProjA,minMax(tDatA)=",ubWw3ProjA.shape,minMax(tDatA)
            print "   mnmx: minMax(ubWw3ProjA)=",minMax(ubWw3ProjA)
            print "   uniques: NPY.unique(ubWw3ProjA).shape=",\
                  NPY.unique(ubWw3ProjA[~NPY.isnan(ubWw3ProjA)]).shape
            toPlotA=NPY.copy(ubWw3ProjA)
            
        if "kd490Proj" in tFileS:
            #There are -9999 nulls in this dataset...
            kd490ProjA=NPY.where(tDatA<0.,\
                    NPY.nanmedian(tDatA),NPY.round(tDatA,3))
            #...accepts it but fills nulls
            #...tDatA Usually 0.0....0.46; has nulls=-9999.
            #   Nulls are made to the dataset median
            #     (follows normal practice for RF)

            #Doesn't need denoising ...
            kd490ProjA=NPY.where(NPY.isnan(wdProjA),fNaN,kd490ProjA)

            kd490ProjA=norm4rf(kd490ProjA,4) #...normalize
            print "kd490ProjA,minMax(tDatA)=",kd490ProjA.shape,minMax(tDatA)
            print "   mnmx: minMax(kd490ProjA)=",minMax(kd490ProjA)
            print "   uniques: NPY.unique(kd490ProjA).shape=",\
                  NPY.unique(kd490ProjA[~NPY.isnan(kd490ProjA)]).shape
            toPlotA=NPY.copy(kd490ProjA)
            
        if "chloraProj" in tFileS:
            #There are -9999 nulls in this dataset...
            chloraProjA=NPY.where(tDatA<0.,\
                    NPY.nanmedian(tDatA),NPY.round(tDatA,4))
            #...accepts it but fills nulls
            #...tDatA Usually 0.0....10.38; has nulls=-9999.
            #   Nulls are made to the dataset median
            #     (follows normal practice for RF)

            #Doesn't need denoising ...
            chloraProjA=NPY.where(NPY.isnan(wdProjA),fNaN,chloraProjA)

            chloraProjA=norm4rf(chloraProjA,4) #...normalize
            print "chloraProjA,minMax(tDatA)=",chloraProjA.shape,minMax(tDatA)
            print "   mnmx: minMax(chloraProjA)=",minMax(chloraProjA)
            print "   uniques: NPY.unique(chloraProjA).shape=",\
                  NPY.unique(chloraProjA[~NPY.isnan(chloraProjA)]).shape
            toPlotA=NPY.copy(chloraProjA)

        X_cols.append(featFileS.replace(".asc",""))

        #Display the feature layer as graphic ...
        PLT.imshow(toPlotA,interpolation='none')
        PLT.title(featFileS)
        PLT.colorbar(shrink=0.5)
        PLT.savefig("_Graphics/"+featFileS.replace(".asc",".png"))
        PLT.close()

        #print "tDatA=",tDatA
    # ----------------------------
    
    print;print "Predictor Variables are in !"
    print "-"*40

    return

# =========================================================================
def norm4rf(arrayA,rfLvls):

    #Normalize a classic way, subtract the mean and rescale to between
    #   -1..1, with number of classes as specified by rfLvls

    print;print ">>norm4rf: rfLvls=",rfLvls

    normedA=arrayA-NPY.nanmean(arrayA)
    normedA=normedA/NPY.nanstd(normedA)
    #eg "https://www.geeksforgeeks.org/python-how-and-where-to-apply-feature-scaling/"

    #But also Trim the ends against severe outliers ...
    pcntile99=NPY.percentile(normedA,99.) 
    pcntile01=NPY.percentile(normedA,01.) 
    normedA=NPY.where(normedA>pcntile99,pcntile99,normedA)
    normedA=NPY.where(normedA<pcntile01,pcntile01,normedA)

    if rfLvls==None:
        #Leave the variable as continuous ...
        pass
    else:
        #Collapse to (usually) specified number of values
        #   of equal population ...
        posRangeA=normedA-NPY.nanmin(normedA) #...takes base to zero.
        unitRangeA=posRangeA/NPY.nanmax(posRangeA) #...takes range to uno.
        print minMax(normedA),minMax(posRangeA),minMax(unitRangeA)
        normedA=NPY.round(unitRangeA*float(rfLvls),0)\
                    /float(rfLvls)*NPY.nanmax(posRangeA)+NPY.nanmin(normedA)
        print "minMax(normedA)=",minMax(normedA)
        print NPY.unique(normedA)[:20]

        normedA=NPY.round(normedA,3)
    
    return normedA

# ===================================================================
def setLimsColoursEtc(gridInS):

    global cbarLabelS
    global vMin,vMax

    print;print ">>setLimsColoursEtc: gridInS=",gridInS
    
    #CJJ From Jan 2017: The colours here will NOT be used...

    if gridInS in "rck,gvl,snd,mud,crb":
        cbarLabelS='pcnt'
        vMin,vMax=[0.-0.001,100.+0.001]

    elif gridInS in "ocbn":
        cbarLabelS='pcnt'
        vMin,vMax=[-2.-0.001,+2.+0.001]

    elif gridInS in "grz":
        cbarLabelS='phi'
        vMin,vMax=[-8.-0.001,+12.+0.001]

    elif gridInS in "srt":
        cbarLabelS='phi'
        vMin,vMax=[0.-0.001,+4.+0.001]

    elif gridInS in "por":
        cbarLabelS='pcnt'
        vMin,vMax=[0.-0.001,100.+0.001]

    else:
        raw_input("Bad choice of gridInS: "+gridInS)

    return

# =========================================================================
def setCmaps(gridInS):

    global cMap

    print;print ">>setCmaps: gridInS=",gridInS
    #...gridInS comes in through global

    #Color pallets...
    if gridInS in "rck.gvl.snd.mud.crb.ocbn.grz.srt.por":
        
        colPalletsD={'rck':'Purples',\
                     'gvl':'Reds','snd':'Wistia','mud':'Greens',\
                     'grz':'gist_ncar_r','srt':'jet','por':'terrain_r',\
                     'crb':'coolwarm_r','ocbn':'copper_r'}
        cMap=PLT.cm.get_cmap(colPalletsD[gridInS])
        cMap.set_over('0.7', 1.0)
        cMap.set_under('0.9', 1.0)
        cMap.set_bad('0.3', 1.0) #...ie land.

        print "   Colourmap: cMap=",cMap

    else:
        cMap=PLT.cm.get_cmap('jet')
        cMap.set_over('0.7', 1.0)
        cMap.set_under('0.9', 1.0)
        cMap.set_bad('0.3', 1.0) #...ie land.

        print "   Default Colourmap: cMap=",cMap

    return 

# =================================================================
def minMax(arrayA):

    return NPY.nanmin(arrayA),NPY.nanmax(arrayA)

# =========================================================================



# =========================================================================
# =========================================================================
#__MAIN__:

#This controls the overall program...

global gridInS

#Specify which parameter to process ---------------------
print;print "Choose the grid/s to process...",
kyS=raw_input("e.g., 'gvl,snd,mud' ")

gridInS=kyS.strip().split(",")[0]

#Set the parameter limits, colourmaps, etc ...
foo=setLimsColoursEtc(gridInS)
foo=setCmaps(gridInS)

#Prepare a dump-file of performance -----------------------------------
try:
    rfSpecsF=open(projS+"_rfSpecs.txt","a")
except:
    rfSpecsF=open(projS+"_rfSpecs.txt","w")
rfSpecsF.write("\n\n"+"-"*20+"\n")
rfSpecsF.write("projS="+projS.upper()+"\n")
todayNowS=time.strftime("%Y-%m-%d %H:%M")
rfSpecsF.write("dateTime="+todayNowS+"\n")

#Main processing ---------------------------------------

print;print "="*40,gridInS
rfSpecsF.write("\n"+"gridInS="+gridInS+"\n")

foo=getEnviroVariables()
#...gets the feature layers ready to append to training data

X_train,y_train=makeTrainData()
#...gets the seabed substrates training (sampled) data

foo=rfFeatureSelection(X_train, y_train)

holdBackQ=True
if holdBackQ:
    foo=holdBackTrainData()
    #... if requested, generates a new holdback file with 10%
    #    of the data, and the old file with 90%; random selecttion

#Not operational yet ...
#Uniquify the training data and map grids for efficiency ...
#foo=uniquifyFeatures()

#INSERT YOUR OWN CODE...
foo=performMachineLearning()

#finishing up ...
rfSpecsF.close()

print;print "Finished !"
# =========================================================================
